#ifndef KKTDIRECT_H
#define KKTDIRECT_H

#ifdef __cplusplus
extern "C" {
#endif

//=================================== integer types ===============================

#ifndef USE_LONG_LONG_INT
typedef int KKT_INT;
typedef unsigned int KKT_UINT;
#else
typedef long long int KKT_INT;
typedef unsigned long long int KKT_UINT;
#endif

//=================================== constraints =================================

// Detect which variables are constrained. At the moment this simply marks
// rows without a diagonal entry (implicit zero) or a non-positive diagonal entry
// as constrained, and the rest as not.
// 
// INPUT:
// n is dimension of matrix.
// colstart is a vector of length n+1, telling where in rowindex each column starts.
// rowindex is a vector of length equal to number of nonzeros in matrix, containg row indices for each column.
// value is a vector of the same length containing the numerical entries of the matrix, or 0 if you
//   don't want to pass one in (in which case implicit zero diagonal entries are the only ones detected).
// (note: all arrays and indices are zero-based)
// OUTPUT:
// constrained should be a vector of length n, which on output indicates which nodes are constrained.

void KKT_detect_constraints(KKT_INT n, 
                            const KKT_INT *colstart,
                            const KKT_INT *rowindex,
                            const double *value,
                            char *constrained);

//=================================== minimum degree ==============================

// Perform minimum degree ordering, respecting KKT structure (if it exists)
// i.e. variables with zero diagonal entries are ordered after their neighbours with nonzero diagonal entries.
// The matrix must have symmetric structure - but this is not explicitly checked.
// Note symmetry implies "row" and "column" are interchangeable in the input description.
//
// INPUT:
// n is dimension of matrix.
// colstart is a vector of length n+1, telling where in rowindex each column starts.
// rowindex is a vector of length equal to number of nonzeros in matrix, containg row indices for each column.
// constrained is a vector of length n, indicating if each variable is constrained (must be ordered
//   after unconstrained neighbours).
// safely_handle_dense_nodes nonzero means when an unconstrained node is preclassified as dense and delayed to the
//   end, all of its constrained neighbours will be delayed after it. If the "B" part of the matrix has a dense
//   column, this may result in catastrophic fill-in; on the other hand, leaving it zero might be unsafe unless
//   "B" without those dense columns still has full row-rank.
// OUTPUT:
// ordering is a vector of length n, which will contain the ordering of the unknowns upon return.

void KKT_minimum_degree(KKT_INT n,
                        const KKT_INT *colstart,
                        const KKT_INT *rowindex,
                        const char *constrained,
                        int safely_handle_dense_nodes,
                        KKT_INT *ordering);

//============================== modifying ordering ===============================

// take an existing ordering and modify it to respect constraints
// (if safely_handle_dense_nodes is zero, remove "dense" unconstrained nodes from consideration in the constraint)
void KKT_modify_ordering_to_respect_constraints(KKT_INT n,
                                                const KKT_INT *colstart,
                                                const KKT_INT *rowindex,
                                                const char *constrained,
                                                int safely_handle_dense_nodes,
                                                KKT_INT *ordering);

//================================= postordering ==================================

// Take the matrix and given ordering, return the elimination tree, a postordering of the tree, and
// the inverse permutation.
//
// INPUT:
// n is dimension of matrix.
// colstart is a vector of length n+1, telling where in rowindex each column starts.
// rowindex is a vector of length equal to number of nonzeros in matrix, containg row indices for each column.
// ordering is a vector of length n, containing the integers 0...n-1 in some order, indicating initial elimination order.
// OUTPUT:
// postordering is a vector of length n, a postordering of the elimination tree.
// inverseordering is a vector of length n, the inverse permutation of postordering.
// etreeparent is a vector of length n, describing the elimination tree (for each node other than a root its parent; -1 for roots)

void KKT_postorder(KKT_INT n, 
                   const KKT_INT *colstart,
                   const KKT_INT *rowindex,
                   const KKT_INT *ordering,
                   KKT_INT *postordering,
                   KKT_INT *inverseordering,
                   KKT_INT *etreeparent);

//============================== symbolic factorization (point) ===================

void KKT_symbolically_factor_uplooking(KKT_INT n,
                                       const KKT_INT *colstart,
                                       const KKT_INT *rowindex,
                                       const KKT_INT *postordering,
                                       const KKT_INT *inverseordering,
                                       const KKT_INT *etreeparent,
                                       KKT_INT *factorcolstart);

//=========================== numerical factorization (point) ======================

// returns 0 if there was a problem (zero or very small pivot) and nonzero otherwise
int KKT_numerically_factor_uplooking(KKT_INT n,
                                     const KKT_INT *colstart,
                                     const KKT_INT *rowindex,
                                     const double *value,
                                     const KKT_INT *postordering,
                                     const KKT_INT *inverseordering,
                                     const KKT_INT *etreeparent,
                                     const KKT_INT *factorcolstart,
                                     KKT_INT *factorrowindex,
                                     double *factorvalue,
                                     double *diagvalue);

//=========================== numerical solution (point) ===========================

void KKT_solve_with_pointwise_factorization(KKT_INT n,
                                            const KKT_INT *postordering,
                                            const KKT_INT *factorcolstart,
                                            const KKT_INT *factorrowindex,
                                            const double *factorvalue,
                                            const double *diagvalue,
                                            const double *righthandside,
                                            double *solution);

//================= symbolic factorization (signed Cholesky, point) ================

void KKT_symbolically_factor_signed_cholesky(KKT_INT n,
                                             const KKT_INT *colstart,
                                             const KKT_INT *rowindex,
                                             const KKT_INT *postordering,
                                             const KKT_INT *inverseordering,
                                             const KKT_INT *etreeparent,
                                             KKT_INT *factorcolstart);

//================= numerical factorization (signed Cholesky, point) ==============

// returns 0 if there was a problem (zero or very small pivot) and nonzero otherwise
int KKT_numerically_factor_signed_cholesky(KKT_INT n,
                                           const KKT_INT *colstart,
                                           const KKT_INT *rowindex,
                                           const double *value,
                                           const char *constrained,
                                           const KKT_INT *postordering,
                                           const KKT_INT *inverseordering,
                                           const KKT_INT *etreeparent,
                                           const KKT_INT *factorcolstart,
                                           KKT_INT *factorrowindex,
                                           double *factorvalue);

//============== numerical solution (signed Cholesky, point) ===================

void KKT_solve_with_signed_cholesky(KKT_INT n,
                                    const char *constrained,
                                    const KKT_INT *postordering,
                                    const KKT_INT *factorcolstart,
                                    const KKT_INT *factorrowindex,
                                    const double *factorvalue,
                                    const double *righthandside,
                                    double *solution);

//=========================== symbolic factorization (supernodal) ==================

// First stage of supernodal symbolic analysis: determines number of supernodes, where each begins,
// and start of each supercolumn.
// numsupernodes should be a pointer to an KKT_INT (which will get the number of supernodes on output)
// supernodestart, supercolstart, factorcolstart, and diagblockstart should all be arrays of length n+1
// (but only the first *numsupernodes+1 entries will be filled in on output)
// supernodemap should be an array of length n (and will be populated with the map from a postordered node to its containing supernode)
void KKT_symbolically_factor_supernodal_initial(KKT_INT n,
                                                const KKT_INT *colstart,
                                                const KKT_INT *rowindex,
                                                const KKT_INT *postordering,
                                                const KKT_INT *inverseordering,
                                                const KKT_INT *etreeparent,
                                                KKT_INT max_supernode_size,
                                                KKT_INT *numsupernodes,
                                                KKT_INT *supernodestart,
                                                KKT_INT *supernodemap,
                                                KKT_INT *supercolstart,
                                                KKT_INT *factorcolstart,
                                                KKT_INT *diagblockstart);

// Second stage: fill in the detailed structure of the factor.
// superrowindex should be a vector of length supercolstart[numsupernodes]
void KKT_symbolically_factor_supernodal_final(KKT_INT n,
                                              const KKT_INT *colstart,
                                              const KKT_INT *rowindex,
                                              const KKT_INT *postordering,
                                              const KKT_INT *inverseordering,
                                              const KKT_INT *etreeparent,
                                              KKT_INT numsupernodes,
                                              const KKT_INT *supernodestart,
                                              const KKT_INT *supernodemap,
                                              const KKT_INT *supercolstart,
                                              KKT_INT *superrowindex);

//======================= numerical factorization (supernodal) ================================

// returns 0 if there was a problem and nonzero otherwise
int KKT_numerically_factor_leftlooking_supernodal(KKT_INT n,
                                                  const KKT_INT *colstart,
                                                  const KKT_INT *rowindex,
                                                  const double *value,
                                                  const KKT_INT *postordering,
                                                  const KKT_INT *inverseordering,
                                                  const KKT_INT *etreeparent,
                                                  // description of supernodes
                                                  KKT_INT numsupernodes,
                                                  const KKT_INT *supernodestart,
                                                  const KKT_INT *supernodemap,
                                                  // description of structure of factor (one supercolumn at a time)
                                                  const KKT_INT *supercolstart,
                                                  const KKT_INT *superrowindex,
                                                  // description of values in factor
                                                  const KKT_INT *factorcolstart,
                                                  double *factorvalue,
                                                  // description of values in diagonal
                                                  const KKT_INT *diagblockstart,
                                                  double *diagvalue,
                                                  // extra stuff for pivoting within diagonal blocks
                                                  KKT_INT *diagpiv);

//======================= numerical solution (supernodal) ================================

void KKT_solve_with_supernodal_factorization(KKT_INT n,
                                             const KKT_INT *postordering,
                                             KKT_INT numsupernodes,
                                             const KKT_INT *supernodestart,
                                             const KKT_INT *supernodemap,
                                             const KKT_INT *supercolstart,
                                             const KKT_INT *superrowindex,
                                             const KKT_INT *factorcolstart,
                                             const double *factorvalue,
                                             const KKT_INT *diagblockstart,
                                             const double *diagvalue,
                                             const KKT_INT *diagpiv,
                                             KKT_INT nrhs,
                                             const double *rhs,
                                             double *solution);

//=========================== symbolic factorization (supernodal signed Cholesky) ==================

void KKT_symbolically_factor_supernodal_initial_for_signed_cholesky(KKT_INT n,
                                                                    const KKT_INT *colstart,
                                                                    const KKT_INT *rowindex,
                                                                    const char *constrained,
                                                                    KKT_INT max_supernode_size,
                                                                    KKT_INT *postordering,
                                                                    KKT_INT *inverseordering,
                                                                    KKT_INT *etreeparent,
                                                                    KKT_INT *numsupernodes,
                                                                    KKT_INT *supernodestart,
                                                                    KKT_INT *supernodemap,
                                                                    KKT_INT *supercolstart,
                                                                    KKT_INT *factorcolstart,
                                                                    KKT_INT *diagblockstart);

// Note: use KKT_symbolically_factor_supernodal_final after this.

//======================= numerical factorization (supernodal signed Cholesky) ================================

// returns 0 if there was a problem and nonzero otherwise
int KKT_numerically_factor_supernodal_signed_cholesky(KKT_INT n,
                                                      const KKT_INT *colstart,
                                                      const KKT_INT *rowindex,
                                                      const double *value,
                                                      const KKT_INT *postordering,
                                                      const KKT_INT *inverseordering,
                                                      const KKT_INT *etreeparent,
                                                      const char *constrained,
                                                      // description of supernodes
                                                      KKT_INT numsupernodes,
                                                      const KKT_INT *supernodestart,
                                                      const KKT_INT *supernodemap,
                                                      // description of structure of factor (one supercolumn at a time)
                                                      const KKT_INT *supercolstart,
                                                      const KKT_INT *superrowindex,
                                                      // description of values in factor
                                                      const KKT_INT *factorcolstart,
                                                      double *factorvalue,
                                                      // description of values in diagonal
                                                      const KKT_INT *diagblockstart,
                                                      double *diagvalue);

//======================= numerical solution (supernodal signed Cholesky) ================================

void KKT_solve_with_supernodal_signed_cholesky(KKT_INT n,
                                               const char *constrained,
                                               const KKT_INT *postordering,
                                               KKT_INT numsupernodes,
                                               const KKT_INT *supernodestart,
                                               const KKT_INT *supernodemap,
                                               const KKT_INT *supercolstart,
                                               const KKT_INT *superrowindex,
                                               const KKT_INT *factorcolstart,
                                               const double *factorvalue,
                                               const KKT_INT *diagblockstart,
                                               const double *diagvalue,
                                               KKT_INT nrhs,
                                               const double *rhs,
                                               double *solution);



//======================================= debugging ==========================
// These routines are probably only useful for debugging purposes.

// returns zero if the candidate ordering violates constraints
int KKT_check_constraints_satisified_by_ordering(KKT_INT n,
                                                 const KKT_INT *colstart,
                                                 const KKT_INT *rowindex,
                                                 const char *constrained,
                                                 const KKT_INT *ordering);

#ifdef __cplusplus
};
#endif

#endif
