#include "kktdirect.h"

// Take an n-by-n saddle point matrix stored in the usual CSC format
// (or CSR - it's symmetric so there should be no difference),
// a right hand side, and then compute and return the solution.
// Returns false if a problem occured.

bool
kktsolve(const int n,
         const int *colstart,
         const int *rowindex,
         const double *value,
         const double *rhs, 
         double *solution)
{
    bool ok=true;

    // order the matrix
    char *constrained=new char[n];
    KKT_detect_constraints(n, colstart, rowindex, value,
                           constrained);
    KKT_INT *order=new KKT_INT[n];
    KKT_minimum_degree(n, colstart, rowindex, constrained, 0,
                       order);

    // compute a postordering and elimination tree
    KKT_INT *postordering=new KKT_INT[n];
    KKT_INT *inverseordering=new KKT_INT[n];
    KKT_INT *etreeparent=new KKT_INT[n];
    KKT_postorder(n, colstart, rowindex, order,
                  postordering, inverseordering, etreeparent);
    delete[] order;

    // now do the factorization and solve!
    const int max_supernode_size=63;

    // first the symbolic factorization
    KKT_INT numsupernodes;
    KKT_INT *supernodestart=new KKT_INT[n+1];
    KKT_INT *supernodemap=new KKT_INT[n];
    KKT_INT *supercolstart=new KKT_INT[n+1];
    KKT_INT *factorcolstart=new KKT_INT[n+1];
    KKT_INT *diagblockstart=new KKT_INT[n+1];
    KKT_symbolically_factor_supernodal_initial_for_signed_cholesky(
            n, colstart, rowindex, constrained, max_supernode_size,
            postordering, inverseordering, etreeparent,
            &numsupernodes, supernodestart, supernodemap, supercolstart,
            factorcolstart, diagblockstart);
    KKT_INT *superrowindex=new KKT_INT[supercolstart[numsupernodes]];
    KKT_symbolically_factor_supernodal_final(
            n, colstart, rowindex, postordering, inverseordering,
            etreeparent, numsupernodes, supernodestart, supernodemap,
            supercolstart, superrowindex);

    // then the numerical factorization
    double *factorvalue=new double[factorcolstart[numsupernodes]];
    double *diagvalue=new double[diagblockstart[numsupernodes]];
    int flag=KKT_numerically_factor_supernodal_signed_cholesky(
            n, colstart, rowindex, value, postordering, inverseordering,
            etreeparent, constrained, numsupernodes, supernodestart,
            supernodemap, supercolstart, superrowindex, factorcolstart,
            factorvalue, diagblockstart, diagvalue);
    if(flag==0){
        ok=false;
        goto finish;
    }

    // and finally the solve
    KKT_solve_with_supernodal_signed_cholesky(
            n, constrained, postordering, numsupernodes, supernodestart,
            supernodemap, supercolstart, superrowindex, factorcolstart,
            factorvalue, diagblockstart, diagvalue, 1, rhs, solution);

    // compute the residual
    double *r=new double[n];
    for(int i=0; i<n; ++i) r[i]=rhs[i];
    for(int i=0; i<n; ++i){
        for(int j=colstart[i]; j<colstart[i+1]; ++j)
            r[rowindex[j]]-=value[j]*solution[i];
    }

    // compute a step of iterative refinement
    double *dx=new double[n];
    KKT_solve_with_supernodal_signed_cholesky(
            n, constrained, postordering, numsupernodes, supernodestart,
            supernodemap, supercolstart, superrowindex, factorcolstart,
            factorvalue, diagblockstart, diagvalue, 1, r, dx);
    for(int i=0; i<n; ++i) solution[i]+=dx[i];

    // and clean up after ourselves
    finish:
    delete[] r;
    delete[] dx;
    delete[] supernodestart;
    delete[] supernodemap;
    delete[] supercolstart;
    delete[] factorcolstart;
    delete[] diagblockstart;
    delete[] superrowindex;
    delete[] factorvalue;
    delete[] diagvalue;
    delete[] constrained;
    delete[] postordering;
    delete[] etreeparent;
    delete[] inverseordering;
    return ok;
}

